/********************************************************************
 * FEMAIN
 ********************************************************************
 *   This is all the FE-PCX code which can still be written in C due
 * to it's lack of need for speed but still must be FE because it is
 * implementation specific.
 ********************************************************************/

#include <malloc.h>
#include <mem.h>
#include "feasm.h"		// Asm funcs specific to this.
#include "fe.h"			// Standard FE header.
#include "universe.h"	// Contains OBJECT definition.

// These defines will have to be made generic:
#define SPRITEW 32
#define SPRITEH 32
#define VIRTW   12
#define VIRTH	9
#define SPRITESZ	1024
#define max(a,b) ((a<b) ? (b) : (a))
#define min(a,b) ((a>b) ? (b) : (a))
#define abs(a)   ((a < 0) ? (-(a)) : (a))
#define POINT_IN(x, y, w, h) \
	((x > 0) && (x < w) && (y > 0) && (y < h))

/********************************************************************
 * File Variables:
 ********************************************************************/
int *xlookup, *ylookup;
char keytable[128];
char transtable[128];
char *virtscreen;
int	 DirtyStack[512];	// More than enough.
int	 DirtyStackPos;
int  RedrawAllFlag = 0;
char map[128][128];
char *tiles[256];
char *rotate[256];		// 256 precompiled rotation codes.
int  *rotatedata[256];	// Edge data for rotation code.
						// Format: Header: width/height, cos/sin
						//		   Element: src off, fract, dst off, dst width.

/********************************************************************
 * Local Function prototypes:
 ********************************************************************/
char *GenerateSpriteCode(int width, int height, int angle);
int  *GenerateEdgeData(int width, int height, int angle);

// This should probably make sure it doesn't add the same square twice.
void AddStack(int *stack, int *stacklen, char *dest, char *source)
{
	stack[*stacklen*2] = (int) dest;
	stack[*stacklen*2 + 1] = (int) source;
	*stacklen += 2;
	}

void DoStack(int *stack, int *stacklen)
{
	int i1;

	for (i1 = 0; i1 < *stacklen;) {
		FEASM_DrawSquare(stack[i1++], stack[i1++]);
		}
	*stacklen = 0;
	}

/********************************************************************
 * FE_Test
 *	Does a cheap test of rotating ability:
 ********************************************************************/
void FE_Test(char *ship, int angle)
{
	char *dest;

	dest = (char *) 0xa0000;
	dest += 100*320 + 160;
	FEASM_DoPrecompiled(dest, ship, rotate[angle]);
//	FEASM_DrawRotated(dest, ship, rotatedata[angle]); // rotatedata[angle]);
	}

/********************************************************************
 * FE_Init
 *	The config file normally would tell us what res etc.  For now,
 * we assume 320x200.
 ********************************************************************/
void FE_Init(char *config) 
{
	int i1;

	if (config)
		i1 = 0;		// Ignore...
	// Create look up tables...
	xlookup = (int *) malloc(1024 * sizeof(int));
	ylookup = (int *) malloc(768 * sizeof(int));
	if (!(ylookup || xlookup)) {
		// Error...  We still need common error recovery!
		}
	// Fill tables:  (This is not currently used until we decide on a scale...
	for (i1 = 0; i1 < 1024; i1++) {
		xlookup[i1] = (int) ((double) i1 / 1024 * 320);
		}
	for (i1 = 0; i1 < 768; i1++) {
		ylookup[i1] = (int) ((double) i1 / 768 * 200);
		}
	// Erase keypress table:
	memset(keytable, 0, 128);
	// Set translation table to 1-1 correspondence.  More on this later.
	for (i1 = 0; i1 < 128; i1++) {
		transtable[i1] = (char) i1;
		}
	// Allocate virtual screen:
	virtscreen = (char *) malloc(SPRITESZ * VIRTW * VIRTH);
	memset(virtscreen, 0, SPRITESZ * VIRTW * VIRTH);
	// Generate prerotated code:
	// Note: code takes 4 mb & only does 746 FPS.
	// Commented out because it's slow.
	for (i1 = 0; i1 < 256; i1++) {
		rotate[i1] = GenerateSpriteCode(SPRITEW, SPRITEH, i1);
		}
	// Generate edge table:
//	for (i1 = 0; i1 < 256; i1++) {
//		rotatedata[i1] = GenerateEdgeData(SPRITEW, SPRITEH, i1);
//		}
	// Grab ints etc:
	FEASM_Init(xlookup, ylookup, keytable);
	}

/********************************************************************
 * FE_Shutdown
 *	Frees memory, kills vectors.
 ********************************************************************/
void FE_Shutdown() 
{
	int i1;

	FEASM_SetGraph(0x3);
	free(xlookup);
	free(ylookup);
	free(virtscreen);
	for (i1 = 0; i1 < 256; i1++) {
		free(rotate[i1]);
		}
	FEASM_ShutDown();
	}

/********************************************************************
 * FE_InitScreen
 *	Sets the screen res etc.
 ********************************************************************/
void FE_InitScreen()
{
	FEASM_SetGraph(0x13);
	
	// Fill dirty stack with entire screen:
	RedrawAllFlag = 1;
	}

/********************************************************************
 * FE_CompileSprite
 *   Takes a 128x128x256c (No colour look up for now...)
 * image & converts it to our own internal size.
 ********************************************************************/
char *FE_CompileSprite(char *source)
{
	char *dest, *temp;
	int	  i1, i2;

	dest = (char *) malloc(SPRITEW * SPRITEH);
	if (!dest) {
		return(dest);
		}
	temp = dest;
	for (i1 = 0; i1 < SPRITEH; i1++) {
		for (i2 = 0; i2 < SPRITEW; i2++) {
			*temp++ = *source;
			source += 4;
			}
		source += 128*3;	// get to next line...
		}
	return(dest);
	}

/********************************************************************
 * FE_IsKey
 *	Tells whether a given key code has been pressed.
 * Basically, looks up in table of key codes for the scan code requested,
 * and checks if that scan code has been pressed.
 ********************************************************************/
int FE_IsKey(char key)
{
	return(keytable[transtable[key]]);
	}

/********************************************************************
 * FE_RedrawScreen
 *	This has yet to be finalized.  Takes a linked list of objects to
 * read (And a rather pathetic object definition to boot)
 ********************************************************************/
void FE_RedrawScreen(OBJECT *first, OBJECT *you, short *ticks)
{
	int nX, nY, i1, i2;
	int offx, offy;
	int Stack[512], StackPos = 0;
	int	dest;
	OBJECT *cur;
	static int oX = -1, oY = -1;

	// First, clear dirty stack:
	DoStack(DirtyStack, &DirtyStackPos);
	
	// Find new upper left...
	// This must be changed to accomadate actual Universe code
	// (Which is still being debated :>
	nX = UNFIX(you->X) / SPRITEW;
	offx = UNFIX(you->X) & (SPRITEW - 1);
	nY = UNFIX(you->Y) / SPRITEH;
	offy = UNFIX(you->Y) & (SPRITEH - 1);
	nX -= VIRTW / 2;		// Get upper left as opposed to centre.
	nY -= VIRTH / 2;
	if (oX == -1) oX = nX;
	if (oY == -1) oY = nY;
	if ((nX != oX) || (nY != oY) || RedrawAllFlag) {
		// move virtual screen:
		if (RedrawAllFlag) {
			// Redraw entire screen (moved to far...)
			RedrawAllFlag = 0;
			oX = nX;
			oY = nY;
			for (i1 = 0; i1 < VIRTH; i1++) {
				for (i2 = 0; i2 < VIRTW; i2++) {
					AddStack(Stack, &StackPos, 
							&virtscreen[i1*SPRITESZ*VIRTW + i2*SPRITEW],
							 tiles[map[oY+i1][oX+i2]]);
					}
				}
			DoStack(Stack, &StackPos);
			}
		else {
			// Do a partial redraw...
			if (oX < nX) {		// Scroll to left.
				while (oX < nX) {
					// Copy screen to left.
					memcpy(virtscreen, &virtscreen[SPRITEW],
							VIRTW*VIRTH*SPRITESZ - SPRITEW);
					// Redraw right edge:
					dest = (int) &virtscreen[VIRTW*SPRITEW - SPRITEW];
					for (i1 = oY; i1 < oY + VIRTH; i1++) {
						AddStack(Stack, &StackPos, (char *) dest, tiles[map[i1][oX+VIRTW]]);
						dest += VIRTW*SPRITESZ;
						}
					oX++;
					DoStack(Stack, &StackPos);
					}
				}
			else {				// Scroll to right.
				while (oX > nX) {
					// Copy screen to right:
					memmove(&virtscreen[SPRITEW], virtscreen,
							 VIRTW*VIRTH*SPRITESZ - SPRITEW);
					dest = (int) virtscreen;
					for (i1 = oY; i1 < oY + VIRTH; i1++) {
						AddStack(Stack, &StackPos, (char *) dest, tiles[map[i1][oX-1]]);
						dest += VIRTW*SPRITESZ;
						}
					oX--;
					DoStack(Stack, &StackPos);
					}
				}
			// Now, adjust Y position:
			if (oY < nY) {		// Scroll up...
				while (oY < nY) {
					memcpy(virtscreen, &virtscreen[SPRITESZ*VIRTW],
						   VIRTW*VIRTH*SPRITESZ - VIRTW*SPRITESZ);
					dest = (int) &virtscreen[(VIRTH-1)*VIRTW*SPRITESZ];
					for (i1 = oX; i1 < oX + VIRTW; i1++) {
						AddStack(Stack, &StackPos, (char *) dest, tiles[map[oY+VIRTH][i1]]);
						dest += SPRITEW;
						}
					oY++;
					DoStack(Stack, &StackPos);
					}
				}
			else {				// Scroll Down...
				while (oY > nY) {
					memmove(&virtscreen[SPRITESZ*VIRTW], virtscreen,
							 VIRTW*VIRTH*SPRITESZ - VIRTW*SPRITESZ);
					dest = (int) virtscreen;
					for (i1 = oX; i1 < oX + VIRTW; i1++) {
						AddStack(Stack, &StackPos, (char *) dest, tiles[map[oY-1][i1]]);
						dest += SPRITEW;
						}
					oY--;
					DoStack(Stack, &StackPos);
					}
				}
			}		// end do partial redraw
		}			// End if nx, ny != sx, sy.
	// We've now redrawn all the floor and (sX, oY) == (nX, nY).
	// Now, draw all the ships
	for (cur = first; cur; cur = cur->next) {
		// Get position of this object:
		nX = UNFIX(cur->X);
		nY = UNFIX(cur->Y);
		// Get where on screen in terms of raw Sprites...
		i1 = (nX / SPRITEW) - oX;
		i2 = (nY / SPRITEH) - oY;
		if ((i1 > 1) && (i2 > 1) && (i1 < VIRTW - 1) && (i2 < VIRTH - 1)) {
			// This is a valid square...  Add to dirty stack
			AddStack(DirtyStack, &DirtyStackPos, 
					 &virtscreen[i2*VIRTW*SPRITESZ + i1*VIRTW],
					  tiles[map[nY/SPRITEH][nX/SPRITEW]]);
			AddStack(DirtyStack, &DirtyStackPos, 
					 &virtscreen[i2*VIRTW*SPRITESZ + i1*VIRTW + VIRTW],
					  tiles[map[nY/SPRITEH][nX/SPRITEW + 1]]);
			AddStack(DirtyStack, &DirtyStackPos, 
					 &virtscreen[(i2+1)*VIRTW*SPRITESZ + i1*VIRTW],
					  tiles[map[nY/SPRITEH + 1][nX/SPRITEW]]);
			AddStack(DirtyStack, &DirtyStackPos, 
					 &virtscreen[(i2+1)*VIRTW*SPRITESZ + (i1+1)*VIRTW],
					  tiles[map[nY/SPRITEH + 1][nX/SPRITEW + 1]]);
			// Dirty squares dealt with...  Now, blit our noble ship.
			FEASM_DoPrecompiled(&virtscreen[(nY - oY*SPRITEH)*VIRTW*SPRITEW +
											  (nX - oX*SPRITEW)],
								  cur->data,
								  rotate[cur->angle]);
			}
		}
	// Okay, backdrop drawn, ships blitted, now actually blit the screen:
	FEASM_BlitScreen(0xa0000, (int) &virtscreen[offy*VIRTW*SPRITEW + offx], 
					VIRTW*SPRITEW-320);
	// All drawn...  Normally retrieve tick count, but that timer not yet
	// added.
	*ticks = 0;
	}

/********************************************************************
 * GenerateSpriteCode
 *	This will generate the proper precompiled code to draw a sprite.
 * Instructions are written in a naive ordering, should be rearranged for
 * proper pairing on 586+.
 ********************************************************************/
char *GenerateSpriteCode(int width, int height, int angle)
{
	int dxx, dxy, dyx, dyy;		// Deltas, (dest:source)
	int sx, sy;					// Source position: 16:16.
	int osx, osy;				// Source position: 16:16, original
	int dx, dy;					// destination position: normal.
	int dw, dh;					// destination width/height.
	int i1, i2;					// Scratch.
	char *code, *ret;
	int codelen = 0;
	int edioff = 0;				// Current edi offset.
	int binit = 0;

	code = (char *) malloc(64000);	// Temporary buffer.
	if (!code)
		return(code);
// Okay, first, find destination width/height.
	i1 = width;				// Remember: we draw from centre!!!
	i2 = height;
	dxx = UN_cos(-angle);
	dxy = UN_sin(-angle);
	dyx = dxy;
	dyy = -dxx;
	dw = (abs(i1*dxx) + abs(i2*dxy)) >> 16;
	dh = (abs(i1*dyx) + abs(i2*dyy)) >> 16;		// Find widths of dest.
	// Ensure dw is a multiple of 4:
	dw = (dw + 3) & (~3L);
// Init sx, sy.
	sx = -dw/2 * UN_cos(-angle) - dh/2 * UN_sin(-angle) + (width/2 * 65536);
	sy = -dw/2 * UN_sin(-angle) + dh/2 * UN_cos(-angle) + (height/2 * 65536);
	for (dy = -dh/2; dy < dh/2; dy++) {
		osx = sx;
		osy = sy;
		// Reset edi for current line:
//		i1 = dy * SPRITEW * VIRTW - dw/2;
		i1 = dy * 320 - dw/2;
		// Write add edi, i1-edioff:
		code[codelen++] = 0x81;
		code[codelen++] = 0xc7;
		*((int *) (code + codelen)) = i1 - edioff;
		edioff = i1;
		codelen += 4;
		for (dx = 0; dx < dw; dx += 4) {
			binit = 0;			// Has this dword been initialized???
			if (POINT_IN((sx>>16), (sy>>16), width, height)) {
				if (!binit) {
					// Init this dword:
					// xor eax, eax; xor edx, edx
					binit = 1;
					code[codelen++] = 0x31;
					code[codelen++] = 0xc0;
					code[codelen++] = 0x31;
					code[codelen++] = 0xd2;
					}
				// Now, load byte:
				// mov al, [esi+off]; sub dl, al; sbb dl, dl.
				code[codelen++] = 0x8a;
				code[codelen++] = 0x86;
				*((int *)(code + codelen)) = (sy >> 16) * SPRITEW + (sx >> 16);
				codelen += 4;
				code[codelen++] = 0x2a;
				code[codelen++] = 0xc2;
				code[codelen++] = 0x1a;
				code[codelen++] = 0xd2;
				}
			sx += dxx;
			sy += dxy;
			if (POINT_IN((sx>>16), (sy>>16), width, height)) {
				if (!binit) {
					// Init this dword:
					// xor eax, eax; xor edx, edx
					binit = 1;
					code[codelen++] = 0x31;
					code[codelen++] = 0xc0;
					code[codelen++] = 0x31;
					code[codelen++] = 0xd2;
					}
				// Now, load byte:
				// mov ah, [esi+off]; sub dl, ah; sbb dl, dl.
				code[codelen++] = 0x8a;
				code[codelen++] = 0xa6;
				*((int *)(code + codelen)) = (sy >> 16) * SPRITEW + (sx >> 16);
				codelen += 4;
				code[codelen++] = 0x2a;
				code[codelen++] = 0xe2;
				code[codelen++] = 0x1a;
				code[codelen++] = 0xd2;
				}
			sx += dxx;
			sy += dxy;
			if (POINT_IN((sx>>16), (sy>>16), width, height)) {
				if (!binit) {
					// Init this dword:
					// xor eax, eax; xor edx, edx
					binit = 1;
					code[codelen++] = 0x31;
					code[codelen++] = 0xc0;
					code[codelen++] = 0x31;
					code[codelen++] = 0xd2;
					}
				else {
					// Already inited, first rotate: (rol eax, 16; rol edx, 16)
					code[codelen++] = 0xc1;
					code[codelen++] = 0xc0;
					code[codelen++] = 0x10;
					code[codelen++] = 0xc1;
					code[codelen++] = 0xc2;
					code[codelen++] = 0x10;
					}
				// Now, load byte:
				// mov al, [esi+off]; sub dl, al; sbb dl, dl.
				code[codelen++] = 0x8a;
				code[codelen++] = 0x86;
				*((int *)(code + codelen)) = (sy >> 16) * SPRITEW + (sx >> 16);
				codelen += 4;
				code[codelen++] = 0x2a;
				code[codelen++] = 0xc2;
				code[codelen++] = 0x1a;
				code[codelen++] = 0xd2;
				}
			sx += dxx;
			sy += dxy;
			if (POINT_IN((sx>>16), (sy>>16), width, height)) {
				if (!binit) {
					// Init this dword:
					// xor eax, eax; xor edx, edx
					binit = 1;
					code[codelen++] = 0x31;
					code[codelen++] = 0xc0;
					code[codelen++] = 0x31;
					code[codelen++] = 0xd2;
					}
				// Now, load byte:
				// mov ah, [esi+off]; sub dl, ah; sbb dl, dl.
				code[codelen++] = 0x8a;
				code[codelen++] = 0xa6;
				*((int *) (code + codelen)) = (sy >> 16) * SPRITEW + (sx >> 16);
				codelen += 4;
				code[codelen++] = 0x2a;
				code[codelen++] = 0xe2;
				code[codelen++] = 0x1a;
				code[codelen++] = 0xd2;
				}
			sx += dxx;
			sy += dxy;
			if (binit) {
				// At this point, edx has an inverted mask & eax has the proper
				// source values.  Both, however, are rotated by 16 bits.
				// Therefore, remove this rotation & blit:
				// rol eax, 16; rol edx, 16; not edx; and edx, [edi+dx];
				// or eax, edx; mov [edi+dx], eax
				code[codelen++] = 0xc1;
				code[codelen++] = 0xc2;
				code[codelen++] = 0x10;
				code[codelen++] = 0xc1;
				code[codelen++] = 0xc0;
				code[codelen++] = 0x10;
				code[codelen++] = 0xf7;
				code[codelen++] = 0xd2;
				code[codelen++] = 0x23;
				code[codelen++] = 0x57;
				code[codelen++] = (char) dx;
				code[codelen++] = 0x0b;
				code[codelen++] = 0xc2;
				code[codelen++] = 0x89;
				code[codelen++] = 0x47;
				code[codelen++] = (char) dx;
				}		// end if binit.
			}		// End for dx=0..dw
		sx = osx;
		sy = osy;
		sx += dyx;
		sy += dyy;
		}			// End for dy=-dh/2..dh/2
	// Write return statement:
	code[codelen++] = 0xc3;
	if (codelen > 65000) {
		// This should NOT happen!!!
		}

	// Create smaller version to return:
	ret = (char *) malloc(codelen);
	if (!ret) {
		free(code);
		return(NULL);
		}
	memcpy(ret, code, codelen);
	free(code);
	return(ret);
	}

/********************************************************************
 * GenerateEdgeData
 * 	 This will generate the appropriate edge data for the rotate code.
 ********************************************************************/
int *GenerateEdgeData(int width, int height, int angle)
{
	int dxx, dxy, dyx, dyy;		// Deltas, (dest:source)
	int sx, sy;					// Source position: 16:16.
	int osx, osy;				// Source position: 16:16, original
	int dx, dy;					// destination position: normal.
	int dw, dh;					// destination width/height.
	int i1, i2;					// Scratch.
	int *code, *ret;
	int codelen = 0;
	int edioff = 0;				// Current edi offset.
	int binit = 0;

	code = (int *) malloc(64000);	// Temporary buffer.
	if (!code)
		return(code);
// Okay, first, find destination width/height.
	i1 = width;				// Remember: we draw from centre!!!
	i2 = height;
	dxx = UN_cos(-angle);
	dxy = UN_sin(-angle);
	dyx = dxy;
	dyy = -dxx;
	dw = (abs(i1*dxx) + abs(i2*dxy)) >> 16;
	dh = (abs(i1*dyx) + abs(i2*dyy)) >> 16;		// Find widths of dest.
// Write header:
	code[codelen++] = dw;
	code[codelen++] = dh;
	code[codelen++] = dxx;
	code[codelen++] = dxy;		// Cos/Sin values used for stepping over map.
// Init sx, sy.
	osx = -dw/2 * UN_cos(-angle) - dh/2 * UN_sin(-angle) + (width/2 * 65536);
	osy = -dw/2 * UN_sin(-angle) + dh/2 * UN_cos(-angle) + (height/2 * 65536);
	for (dy = -dh/2; dy < -dh/2 + dh; dy++) {
		sx = osx;
		sy = osy;
		for (dx = 0; dx < dw && !POINT_IN((sx>>16), (sy>>16), width, height); dx++) {
			sx += dxx;
			sy += dxy;
			}
		// Write src offset
		code[codelen++] = (sx >> 16);
		// Write fractional source offset: Contains real sy value as well!!!
		code[codelen++] = ((sx & 0xffff) << 16) + ((sy >> 8) & 0xffff);
		// write dest offset:
		code[codelen++] = dx;
		for (; dx < dw && POINT_IN((sx>>16), (sy>>16), width, height); dx++) {
			sx += dxx;
			sy += dxy;
			}
		// write width:
		code[codelen] = dx - code[codelen-1];
		codelen++;
		osx += dyx;
		osy += dyy;
		}			// End for dy=-dh/2..dh/2

	if (codelen > 64000/sizeof(int)) {
		// This should NOT happen!!!
		}

	// Create smaller version to return:
	ret = (int *) malloc(codelen*sizeof(int));
	if (!ret) {
		free(code);
		return(NULL);
		}
	memcpy(ret, code, codelen*sizeof(int));
	free(code);
	return(ret);
	}
